# Metrica Documentation https://github.com/metrica-sports/sample-data/blob/master/documentation/events-definitions.pdf
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Type, Union, Any

from kloppy.domain.models.common import DatasetType
from kloppy.utils import (
    camelcase_to_snakecase,
    removes_suffix,
    docstring_inherit_attributes,
)

from .common import DataRecord, Dataset, Player, Team
from .pitch import Point


class ResultType(Enum):
    @property
    @abstractmethod
    def is_success(self):
        raise NotImplementedError


class ShotResult(ResultType):
    """
    ShotResult

    Attributes:
        GOAL (ShotResult): Shot resulted in a goal
        OFF_TARGET (ShotResult): Shot was off target
        POST (ShotResult): Shot hit the post
        BLOCKED (ShotResult): Shot was blocked by another player
        SAVED (ShotResult): Shot was saved by the keeper
    """

    GOAL = "GOAL"
    OFF_TARGET = "OFF_TARGET"
    POST = "POST"
    BLOCKED = "BLOCKED"
    SAVED = "SAVED"

    @property
    def is_success(self):
        """
        Returns if the shot was a goal
        """
        return self == self.GOAL


class PassResult(ResultType):
    """
    PassResult

    Attributes:
        COMPLETE (PassResult): Complete pass
        INCOMPLETE (PassResult): Incomplete pass (intercepted)
        OUT (PassResult): Ball went out
        OFFSIDE (PassResult): Offside
    """

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"
    OUT = "OUT"
    OFFSIDE = "OFFSIDE"

    @property
    def is_success(self):
        """
        Returns if the pass was complete
        """
        return self == self.COMPLETE


class TakeOnResult(ResultType):
    """
    TakeOnResult

    Attributes:
        COMPLETE (TakeOnResult): Complete take-on
        INCOMPLETE (TakeOnResult): Incomplete take-on
        OUT (TakeOnResult): Ball went out
    """

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"
    OUT = "OUT"

    @property
    def is_success(self):
        """
        Returns if the take-on was complete
        """
        return self == self.COMPLETE


class CarryResult(ResultType):
    """
    CarryResult

    Attributes:
        COMPLETE (CarryResult): Complete carry
        INCOMPLETE (CarryResult): Incomplete carry
    """

    COMPLETE = "COMPLETE"
    INCOMPLETE = "INCOMPLETE"

    @property
    def is_success(self):
        """
        Returns if the carry was complete
        """
        return self == self.COMPLETE


class CardType(Enum):
    """
    CardType

    Attributes:
        FIRST_YELLOW (CardType): First yellow card
        SECOND_YELLOW (CardType): Second yellow card
        RED (CardType): Red card
    """

    FIRST_YELLOW = "FIRST_YELLOW"
    SECOND_YELLOW = "SECOND_YELLOW"
    RED = "RED"


class EventType(Enum):
    """
    Attributes:
        GENERIC (EventType): Unrecognised event type
        PASS (EventType):
        SHOT (EventType):
        TAKE_ON (EventType):
        CARRY (EventType):
        SUBSTITUTION (EventType):
        CARD (EventType):
        PLAYER_ON (EventType):
        PLAYER_OFF (EventType):
        RECOVERY (EventType):
        BALL_OUT (EventType):
        FOUL_COMMITTED (EventType):
    """

    GENERIC = "generic"

    PASS = "PASS"
    SHOT = "SHOT"
    TAKE_ON = "TAKE_ON"
    CARRY = "CARRY"
    SUBSTITUTION = "SUBSTITUTION"
    CARD = "CARD"
    PLAYER_ON = "PLAYER_ON"
    PLAYER_OFF = "PLAYER_OFF"
    RECOVERY = "RECOVERY"
    BALL_OUT = "BALL_OUT"
    FOUL_COMMITTED = "FOUL_COMMITTED"

    def __repr__(self):
        return self.value


@dataclass
class Qualifier(ABC):
    value: None

    @abstractmethod
    def to_dict(self):
        """
        Return the qualifier as a dict
        """
        pass

    @property
    def name(self):
        return camelcase_to_snakecase(
            removes_suffix(type(self).__name__, "Qualifier")
        )


@dataclass
class BoolQualifier(Qualifier, ABC):
    value: bool

    def to_dict(self):
        return {f"is_{self.name}": self.value}


class EnumQualifier(Qualifier, ABC):
    value: Enum

    def to_dict(self):
        return {f"{self.name}_type": self.value.value}


class SetPieceType(Enum):
    """
    SetPieceType

    Attributes:
        GOAL_KICK (SetPieceType):
        FREE_KICK (SetPieceType):
        THROW_IN (SetPieceType):
        CORNER_KICK (SetPieceType):
        PENALTY (SetPieceType):
        KICK_OFF (SetPieceType):
    """

    GOAL_KICK = "GOAL_KICK"
    FREE_KICK = "FREE_KICK"
    THROW_IN = "THROW_IN"
    CORNER_KICK = "CORNER_KICK"
    PENALTY = "PENALTY"
    KICK_OFF = "KICK_OFF"


@dataclass
class SetPieceQualifier(EnumQualifier):
    """
    SetPieceQualifier

    Attributes:
        value: Specifies the type of set piece
    """

    value: SetPieceType


class PassType(Enum):
    """
    PassType

    Attributes:
        CROSS (PassType):
        HAND_PASS (PassType):
        HEAD_PASS (PassType):
        HIGH_PASS (PassType):
        LAUNCH (PassType):
        SIMPLE_PASS (PassType):
        SMART_PASS (PassType):
    """

    CROSS = "CROSS"
    HAND_PASS = "HAND_PASS"
    HEAD_PASS = "HEAD_PASS"
    HIGH_PASS = "HIGH_PASS"
    LAUNCH = "LAUNCH"
    SIMPLE_PASS = "SIMPLE_PASS"
    SMART_PASS = "SMART_PASS"


@dataclass
class PassQualifier(EnumQualifier):
    value: PassType


class BodyPart(Enum):
    """
    BodyPart

    Attributes:
        RIGHT_FOOT (BodyPart): Pass or Shot with right foot, save with right foot (for goalkeepers).
        LEFT_FOOT (BodyPart): Pass or Shot with leftt foot, save with left foot (for goalkeepers).
        HEAD (BodyPart): Pass or Shot with head, save with head (for goalkeepers).
        BOTH_HANDS (BodyPart): Goalkeeper only. Save with both hands.
        CHEST (BodyPart): Goalkeeper only. Save with chest.
        LEFT_HAND (BodyPart): Goalkeeper only. Save with left hand.
        RIGHT_HAND (BodyPart): Goalkeeper only. Save with right hand.
        DROP_KICK (BodyPart): Pass is a keeper drop kick.
        KEEPER_ARM (BodyPart): Pass thrown from keepers hands.
        OTHER (BodyPart): Other body part (chest, back, etc.), for Pass and Shot.
        NO_TOUCH (BodyPart): Pass only. A player deliberately let the pass go past him
                             instead of receiving it to deliver to a teammate behind him.
                             (Also known as a "dummy").
    """

    RIGHT_FOOT = "RIGHT_FOOT"
    LEFT_FOOT = "LEFT_FOOT"
    HEAD = "HEAD"

    BOTH_HANDS = "BOTH_HANDS"
    CHEST = "CHEST"
    LEFT_HAND = "LEFT_HAND"
    RIGHT_HAND = "RIGHT_HAND"
    DROP_KICK = "DROP_KICK"
    KEEPER_ARM = "KEEPER_ARM"
    OTHER = "OTHER"
    NO_TOUCH = "NO_TOUCH"


@dataclass
class BodyPartQualifier(EnumQualifier):
    value: BodyPart


class GoalkeeperAction(Enum):
    REFLEX = "REFLEX"
    SAVE_ATTEMPT = "SAVE_ATTEMPT"


@dataclass
class GoalkeeperActionQualifier(EnumQualifier):
    value: GoalkeeperAction


@dataclass
class CounterAttackQualifier(BoolQualifier):
    pass


@dataclass
@docstring_inherit_attributes(DataRecord)
class Event(DataRecord, ABC):
    """
    Abstract event baseclass. All other event classes inherit from this class.

    Attributes:
        event_id: identifier given by provider
        team: See [`Team`][kloppy.domain.models.common.Team]
        player: See [`Player`][kloppy.domain.models.common.Player]
        coordinates: Coordinates where event happened. See [`Point`][kloppy.domain.models.pitch.Point]
        raw_event: Dict
        state: Dict[str, Any]
        qualifiers: See [`Qualifier`][kloppy.domain.models.event.Qualifier]
    """

    event_id: str
    team: Team
    player: Player
    coordinates: Point

    result: Union[ResultType, None]

    raw_event: Dict
    state: Dict[str, Any]

    qualifiers: List[Qualifier]

    @property
    @abstractmethod
    def event_type(self) -> EventType:
        raise NotImplementedError

    @property
    @abstractmethod
    def event_name(self) -> str:
        raise NotImplementedError

    @classmethod
    def create(cls, **kwargs):
        return cls(**kwargs, state={})

    def get_qualifier_value(self, qualifier_type: Type[Qualifier]):
        """
        Returns the Qualifier of a certain type, or None if qualifier is not present.

        Arguments:
            qualifier_type: one of the following QualifierTypes: [`SetPieceQualifier`][kloppy.domain.models.event.SetPieceQualifier]
                [`BodyPartQualifier`][kloppy.domain.models.event.BodyPartQualifier] [`PassQualifier`][kloppy.domain.models.event.PassQualifier]

        Examples:
            >>> from kloppy.domain import SetPieceQualifier
            >>> pass_event.get_qualifier_value(SetPieceQualifier)
            <SetPieceType.GOAL_KICK: 'GOAL_KICK'>
        """
        if self.qualifiers:
            for qualifier in self.qualifiers:
                if isinstance(qualifier, qualifier_type):
                    return qualifier.value
        return None


@dataclass
@docstring_inherit_attributes(Event)
class GenericEvent(Event):
    """
    GenericEvent

    Attributes:
        event_type (EventType): `EventType.GENERIC` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"generic"`
    """

    event_type: EventType = EventType.GENERIC
    event_name: str = "generic"


@dataclass
@docstring_inherit_attributes(Event)
class ShotEvent(Event):
    """
    ShotEvent

    Attributes:
        event_type (EventType): `EventType.SHOT` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"shot"`,
        result_coordinates (Point): See [`Point`][kloppy.domain.models.pitch.Point]
        result (ShotResult): See [`ShotResult`][kloppy.domain.models.event.ShotResult]
    """

    result: ShotResult
    result_coordinates: Point = None

    event_type: EventType = EventType.SHOT
    event_name: str = "shot"


@dataclass
@docstring_inherit_attributes(Event)
class PassEvent(Event):
    """
    PassEvent

    Attributes:
        event_type (EventType): `EventType.PASS` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"pass"`
        receive_timestamp (float):
        receiver_coordinates (Point): See [`Point`][kloppy.domain.models.pitch.Point]
        receiver_player (Player): See [`Player`][kloppy.domain.models.common.Player]
        result (PassResult): See [`PassResult`][kloppy.domain.models.event.PassResult]
    """

    receive_timestamp: float
    receiver_player: Player
    receiver_coordinates: Point

    result: PassResult

    event_type: EventType = EventType.PASS
    event_name: str = "pass"


@dataclass
@docstring_inherit_attributes(Event)
class TakeOnEvent(Event):
    """
    TakeOnEvent

    Attributes:
        event_type (EventType): `EventType.TAKE_ON` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"take_on"`,
        result (TakeOnResult): See [`TakeOnResult`][kloppy.domain.models.event.TakeOnResult]
    """

    result: TakeOnResult

    event_type: EventType = EventType.TAKE_ON
    event_name: str = "take_on"


@dataclass
@docstring_inherit_attributes(Event)
class CarryEvent(Event):
    """
    CarryEvent

    Attributes:
        event_type (EventType): `EventType.CARRY` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"carry"`,
        end_timestamp (float):
        end_coordinates (Point): See [`Point`][kloppy.domain.models.pitch.Point]
        result (CarryResult): See [`CarryResult`][kloppy.domain.models.event.CarryResult]
    """

    end_timestamp: float
    end_coordinates: Point

    result: CarryResult

    event_type: EventType = EventType.CARRY
    event_name: str = "carry"


@dataclass
@docstring_inherit_attributes(Event)
class SubstitutionEvent(Event):
    """
    SubstitutionEvent

    Attributes:
        event_type (EventType): `EventType.SUBSTITUTION` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"substitution"`,
        replacement_player (Player): See [`Player`][kloppy.domain.models.common.Player]
    """

    replacement_player: Player

    event_type: EventType = EventType.SUBSTITUTION
    event_name: str = "substitution"


@dataclass
@docstring_inherit_attributes(Event)
class PlayerOffEvent(Event):
    """
    PlayerOffEvent

    Attributes:
        event_type (EventType): `EventType.PLAYER_OFF` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"player_off"`
    """

    event_type: EventType = EventType.PLAYER_OFF
    event_name: str = "player_off"


@dataclass
@docstring_inherit_attributes(Event)
class PlayerOnEvent(Event):
    """
    PlayerOnEvent

    Attributes:
        event_type (EventType): `EventType.PLAYER_ON` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"player_on"`
    """

    event_type: EventType = EventType.PLAYER_ON
    event_name: str = "player_on"


@dataclass
@docstring_inherit_attributes(Event)
class CardEvent(Event):
    """
    CardEvent

    Attributes:
        event_type (EventType): `EventType.CARD` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): `"card"`
        card_type: See [`CardType`][kloppy.domain.models.event.CardType]
    """

    card_type: CardType

    event_type: EventType = EventType.CARD
    event_name: str = "card"


@dataclass
@docstring_inherit_attributes(Event)
class RecoveryEvent(Event):
    """
    RecoveryEvent

    Attributes:
        event_type (EventType): `EventType.RECOVERY` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): "recovery"
    """

    event_type: EventType = EventType.RECOVERY
    event_name: str = "recovery"


@dataclass
@docstring_inherit_attributes(Event)
class BallOutEvent(Event):
    """
    BallOutEvent

    Attributes:
        event_type (EventType): `EventType.BALL_OUT` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): "ball_out"
    """

    event_type: EventType = EventType.BALL_OUT
    event_name: str = "ball_out"


@dataclass
@docstring_inherit_attributes(Event)
class FoulCommittedEvent(Event):
    """
    FoulCommittedEvent

    Attributes:
        event_type (EventType): `EventType.FOUL_COMMITTED` (See [`EventType`][kloppy.domain.models.event.EventType])
        event_name (str): "foul_committed"
    """

    event_type: EventType = EventType.FOUL_COMMITTED
    event_name: str = "foul_committed"


@dataclass
class EventDataset(Dataset):
    """
    EventDataset

    Attributes:
        metadata: See [`Metadata`][kloppy.domain.models.common.Metadata]
        records (List[Event]): See [`Event`][kloppy.domain.models.event.Event]
        dataset_type: `DatasetType.EVENT` (See [`DatasetType`][kloppy.domain.models.common.DatasetType])
        events: alias for `records`
    """

    records: List[
        Union[
            GenericEvent,
            ShotEvent,
            PassEvent,
            TakeOnEvent,
            CarryEvent,
            SubstitutionEvent,
            PlayerOffEvent,
            PlayerOnEvent,
            CardEvent,
        ]
    ]

    dataset_type: DatasetType = DatasetType.EVENT

    @property
    def events(self):
        return self.records

    def add_state(self, *args, **kwargs):
        """
        See [add_state][kloppy.domain.services.state_builder.add_state]
        """
        from kloppy import add_state

        return add_state(self, *args, **kwargs)


__all__ = [
    "ResultType",
    "EventType",
    "ShotResult",
    "PassResult",
    "TakeOnResult",
    "CarryResult",
    "Event",
    "GenericEvent",
    "ShotEvent",
    "PassEvent",
    "TakeOnEvent",
    "CarryEvent",
    "SubstitutionEvent",
    "PlayerOnEvent",
    "PlayerOffEvent",
    "CardEvent",
    "CardType",
    "EventDataset",
    "RecoveryEvent",
    "FoulCommittedEvent",
    "BallOutEvent",
    "SetPieceType",
    "Qualifier",
    "SetPieceQualifier",
    "PassQualifier",
    "PassType",
    "BodyPart",
    "BodyPartQualifier",
    "GoalkeeperAction",
    "GoalkeeperActionQualifier",
    "CounterAttackQualifier",
]



import os
from itertools import groupby

from kloppy import StatsBombSerializer, add_state, to_pandas
from kloppy.domain import EventType, Event, EventDataset
from kloppy.domain.services.state_builder.builder import StateBuilder, T
from kloppy.utils import performance_logging


class TestStateBuilder:
    def _load_dataset(self, base_filename="statsbomb", options=None):
        base_dir = os.path.dirname(__file__)

        serializer = StatsBombSerializer()

        with open(
            f"{base_dir}/files/{base_filename}_lineup.json", "rb"
        ) as lineup_data, open(
            f"{base_dir}/files/{base_filename}_event.json", "rb"
        ) as event_data:
            dataset = serializer.deserialize(
                inputs={"lineup_data": lineup_data, "event_data": event_data},
                options=options,
            )
            return dataset

    def test_score_state_builder(self):
        dataset = self._load_dataset()

        with performance_logging("add_state"):
            dataset_with_state = add_state(dataset, ["score"])

        events_per_score = {}
        for score, events in groupby(
            dataset_with_state.events, lambda event: event.state["score"]
        ):
            events = list(events)
            events_per_score[str(score)] = len(events)

        assert events_per_score == {
            "0-0": 2897,
            "1-0": 717,
            "2-0": 405,
            "3-0": 3,
        }

    def test_sequence_state_builder(self):
        dataset = self._load_dataset()

        with performance_logging("add_state"):
            dataset_with_state = add_state(dataset, ["sequence"])

        events_per_sequence = {}
        for sequence_id, events in groupby(
            dataset_with_state.events,
            lambda event: event.state["sequence"].sequence_id,
        ):
            events = list(events)
            events_per_sequence[sequence_id] = len(events)

        assert events_per_sequence[0] == 4
        assert events_per_sequence[50] == 14

    def test_lineup_state_builder(self):
        dataset = self._load_dataset("statsbomb_15986")

        with performance_logging("add_state"):
            dataset_with_state = add_state(dataset, ["lineup"])

        last_events = []
        for lineup, events in groupby(
            dataset_with_state.events, lambda event: event.state["lineup"]
        ):
            events = list(events)
            # inspect last event which changed the lineup
            last_events.append((events[-1].event_type, len(lineup.players)))

        assert last_events == [
            (EventType.CARD, 22),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.SUBSTITUTION, 21),
            (EventType.GENERIC, 21),
        ]

    def test_register_custom_builder(self):
        class CustomStateBuilder(StateBuilder):
            def initial_state(self, dataset: EventDataset) -> int:
                return 0

            def reduce_before(self, state: int, event: Event) -> int:
                return state + 1

            def reduce_after(self, state: int, event: Event) -> int:
                return state + 1

        dataset = self._load_dataset("statsbomb_15986")

        with performance_logging("add_state"):
            dataset_with_state = add_state(dataset, ["custom"])

        assert dataset_with_state.events[0].state["custom"] == 1
        assert dataset_with_state.events[1].state["custom"] == 3
        assert dataset_with_state.events[2].state["custom"] == 5
        assert dataset_with_state.events[3].state["custom"] == 7



from dataclasses import dataclass, field
from typing import List, Dict

from kloppy.domain.models.common import DatasetType

from .common import Dataset, DataRecord
from ...utils import docstring_inherit_attributes


@dataclass
@docstring_inherit_attributes(DataRecord)
class Code(DataRecord):
    """
    Single code

    Attributes:
        id: identifier provided by the coding software
        code: string describing the code
        end_timestamp: float
        labels: Text labels describing this code instance

    """

    code_id: str
    code: str
    end_timestamp: float
    labels: Dict[str, str] = field(default_factory=dict)

    @property
    def start_timestamp(self):
        return self.timestamp


@dataclass
class CodeDataset(Dataset):
    records: List[Code]

    dataset_type: DatasetType = DatasetType.CODE

    @property
    def codes(self):
        return self.records


__all__ = ["Code", "CodeDataset"]



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterator,
    Mapping,
    Sequence,
    Text,
    Tuple,
    TypeVar,
    Union,
    Dict,
)

if TYPE_CHECKING:
    from .ast import Final

# Tokens type
Tok = TypeVar("Tok")

# Output type
Out = TypeVar("Out")


@dataclass
class _TrailItem(Generic[Out]):
    """
    Internal intermediate object that will represent a step in the matching
    process along with the edge metadata.

    Let's just mention that the matched item is not the input token but the
    output of the matcher for this token.
    """

    item: Out
    data: Dict[Text, Sequence[Text]]

    @property
    def _comparable(self):
        """
        Returns a signature tuple that can be used for comparison
        """

        return self.item, tuple(sorted(self.data.items()))

    def __lt__(self, other):
        """
        Sortability for de-duplication purposes
        """

        if not isinstance(other, _TrailItem):
            raise TypeError

        return self._comparable < other._comparable


class Matcher(Generic[Tok, Out], metaclass=ABCMeta):
    @abstractmethod
    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        raise NotImplementedError


class Eq(Matcher):
    def __init__(self, ref: Tok):
        self.ref = ref

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        if self.ref == token:
            yield token

    def __repr__(self):
        return f"Eq({self.ref!r})"


class In(Matcher):
    def __init__(self, ref: Sequence[Tok]):
        self.ref = ref

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        if token in self.ref:
            yield token

    def __repr__(self):
        return f"In({self.ref!r})"


class OutOf(Matcher):
    def __init__(self, ref: Tok):
        self.ref = ref

    def __repr__(self):
        return f"OutOf({self.ref!r})"

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        if self.ref in token:
            yield self.ref


class AttributeHasValue(Matcher):
    def __init__(self, attribute: Text, value: Any):
        self.attribute = attribute
        self.value = value

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        if (
            hasattr(token, self.attribute)
            and getattr(token, self.attribute) == self.value
        ):
            yield token

    def __repr__(self):
        return f"AttributeHasValue({self.attribute}={self.value!r}"


class KeyHasValue(Matcher):
    def __init__(self, key: Any, value: Any):
        self.key = key
        self.value = value

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        if (
            isinstance(token, Mapping)
            and self.key in token
            and token[self.key] == self.value
        ):
            yield token

    def __repr__(self):
        return f"KeyHasValue({self.key!r}={self.value!r})"


class Anything(Matcher):
    def __repr__(self):
        return f"Anything()"

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        yield token


class ChrRanges(Matcher[str, str]):
    def __init__(self, *ranges: Tuple[str, str]):
        self.ranges = ranges

    def __repr__(self):
        return f"ChrRanges{self.ranges!r}"

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        for start, stop in self.ranges:
            if ord(start) <= ord(token) <= ord(stop):
                yield token


class Test(Matcher[Tok, Out]):
    """
    Runs an arbitrary test and matches the token as-is if successful
    """

    def __init__(self, test: Callable[[Tok], bool]):
        self.test = test

    def __repr__(self):
        return f"Test({self.test!r}"

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        if self.test(token):
            yield token


class Not(Matcher[Tok, Out]):
    """
    Negates another matcher. If no match is found then the token is returned
    as-is.
    """

    def __init__(self, matcher: Union[Matcher, "Final"]):
        from .ast import Final

        if isinstance(matcher, Final):
            self.matcher: Matcher[Tok, Out] = matcher.statement
        elif isinstance(matcher, Matcher):
            self.matcher: Matcher[Tok, Out] = matcher
        else:
            raise ValueError("Cannot negate this")

    def __repr__(self):
        return f"Not({self.matcher!r})"

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        found = list(self.matcher.match(token, trail))

        if not found:
            yield token


__all__ = [
    "Tok",
    "Out",
    "Matcher",
    "Eq",
    "In",
    "OutOf",
    "AttributeHasValue",
    "KeyHasValue",
    "Anything",
    "ChrRanges",
    "Test",
    "Not",
    "_TrailItem",
]



from typing import Tuple, Dict, List
import logging
from datetime import datetime
import pytz
from lxml import objectify

from kloppy.domain import (
    EventDataset,
    Team,
    Period,
    Point,
    BallState,
    DatasetFlag,
    Orientation,
    PitchDimensions,
    Dimension,
    PassEvent,
    ShotEvent,
    TakeOnEvent,
    CarryEvent,
    GenericEvent,
    PassResult,
    ShotResult,
    TakeOnResult,
    CarryResult,
    EventType,
    Ground,
    Score,
    Provider,
    Metadata,
    Player,
    Position,
    RecoveryEvent,
    BallOutEvent,
    FoulCommittedEvent,
    Qualifier,
    SetPieceQualifier,
    SetPieceType,
)
from kloppy.infra.serializers.event import EventDataSerializer
from kloppy.utils import Readable, performance_logging

logger = logging.getLogger(__name__)

EVENT_TYPE_START_PERIOD = 32
EVENT_TYPE_END_PERIOD = 30

EVENT_TYPE_PASS = 1
EVENT_TYPE_OFFSIDE_PASS = 2
EVENT_TYPE_TAKE_ON = 3
EVENT_TYPE_SHOT_MISS = 13
EVENT_TYPE_SHOT_POST = 14
EVENT_TYPE_SHOT_SAVED = 15
EVENT_TYPE_SHOT_GOAL = 16
EVENT_TYPE_BALL_OUT = 5
EVENT_TYPE_CORNER_AWARDED = 6
EVENT_TYPE_FOUL_COMMITTED = 4
EVENT_TYPE_RECOVERY = 49

BALL_OUT_EVENTS = [EVENT_TYPE_BALL_OUT, EVENT_TYPE_CORNER_AWARDED]

BALL_OWNING_EVENTS = (
    EVENT_TYPE_PASS,
    EVENT_TYPE_OFFSIDE_PASS,
    EVENT_TYPE_TAKE_ON,
    EVENT_TYPE_SHOT_MISS,
    EVENT_TYPE_SHOT_POST,
    EVENT_TYPE_SHOT_SAVED,
    EVENT_TYPE_SHOT_GOAL,
    EVENT_TYPE_RECOVERY,
)

EVENT_QUALIFIER_GOAL_KICK = 124
EVENT_QUALIFIER_FREE_KICK = 5
EVENT_QUALIFIER_THROW_IN = 107
EVENT_QUALIFIER_CORNER_KICK = 6
EVENT_QUALIFIER_PENALTY = 9
EVENT_QUALIFIER_KICK_OFF = 279

event_type_names = {
    1: "pass",
    2: "offside pass",
    3: "take on",
    4: "foul",
    5: "out",
    6: "corner awarded",
    7: "tackle",
    8: "interception",
    9: "turnover",
    10: "save",
    11: "claim",
    12: "clearance",
    13: "miss",
    14: "post",
    15: "attempt saved",
    16: "goal",
    17: "card",
    18: "player off",
    19: "player on",
    20: "player retired",
    21: "player returns",
    22: "player becomes goalkeeper",
    23: "goalkeeper becomes player",
    24: "condition change",
    25: "official change",
    26: "unknown26",
    27: "start delay",
    28: "end delay",
    29: "unknown29",
    30: "end",
    31: "unknown31",
    32: "start",
    33: "unknown33",
    34: "team set up",
    35: "player changed position",
    36: "player changed jersey number",
    37: "collection end",
    38: "temp_goal",
    39: "temp_attempt",
    40: "formation change",
    41: "punch",
    42: "good skill",
    43: "deleted event",
    44: "aerial",
    45: "challenge",
    46: "unknown46",
    47: "rescinded card",
    48: "unknown46",
    49: "ball recovery",
    50: "dispossessed",
    51: "error",
    52: "keeper pick-up",
    53: "cross not claimed",
    54: "smother",
    55: "offside provoked",
    56: "shield ball opp",
    57: "foul throw in",
    58: "penalty faced",
    59: "keeper sweeper",
    60: "chance missed",
    61: "ball touch",
    62: "unknown62",
    63: "temp_save",
    64: "resume",
    65: "contentious referee decision",
    66: "possession data",
    67: "50/50",
    68: "referee drop ball",
    69: "failed to block",
    70: "injury time announcement",
    71: "coach setup",
    72: "caught offside",
    73: "other ball contact",
    74: "blocked pass",
    75: "delayed start",
    76: "early end",
    77: "player off pitch",
}


def _parse_f24_datetime(dt_str: str) -> float:
    return (
        datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%f")
        .replace(tzinfo=pytz.utc)
        .timestamp()
    )


def _parse_pass(raw_qualifiers: Dict[int, str], outcome: int) -> Dict:
    if outcome:
        receiver_coordinates = Point(
            x=float(raw_qualifiers[140]), y=float(raw_qualifiers[141])
        )
        result = PassResult.COMPLETE
    else:
        result = PassResult.INCOMPLETE
        receiver_coordinates = None

    qualifiers = _get_event_qualifiers(raw_qualifiers)

    return dict(
        result=result,
        receiver_coordinates=receiver_coordinates,
        receiver_player=None,
        receive_timestamp=None,
        qualifiers=qualifiers,
    )


def _parse_offside_pass(raw_qualifiers: List) -> Dict:
    qualifiers = _get_event_qualifiers(raw_qualifiers)
    return dict(
        result=PassResult.OFFSIDE,
        receiver_coordinates=None,
        receiver_player=None,
        receive_timestamp=None,
        qualifiers=qualifiers,
    )


def _parse_take_on(outcome: int) -> Dict:
    if outcome:
        result = TakeOnResult.COMPLETE
    else:
        result = TakeOnResult.INCOMPLETE
    return dict(result=result)


def _parse_shot(
    raw_qualifiers: Dict[int, str], type_id: int, coordinates: Point
) -> Dict:
    if type_id == EVENT_TYPE_SHOT_GOAL:
        if 28 in raw_qualifiers:
            coordinates = Point(x=100 - coordinates.x, y=100 - coordinates.y)
        result = ShotResult.GOAL
    else:
        result = None

    qualifiers = _get_event_qualifiers(raw_qualifiers)

    return dict(coordinates=coordinates, result=result, qualifiers=qualifiers)


def _parse_team_players(
    f7_root, team_ref: str
) -> Tuple[str, Dict[str, Dict[str, str]]]:
    matchdata_path = objectify.ObjectPath("SoccerFeed.SoccerDocument")
    team_elms = list(matchdata_path.find(f7_root).iterchildren("Team"))
    for team_elm in team_elms:
        if team_elm.attrib["uID"] == team_ref:
            team_name = str(team_elm.find("Name"))
            players = {
                player_elm.attrib["uID"]: dict(
                    first_name=str(
                        player_elm.find("PersonName").find("First")
                    ),
                    last_name=str(player_elm.find("PersonName").find("Last")),
                )
                for player_elm in team_elm.iterchildren("Player")
            }
            break
    else:
        raise Exception(f"Could not parse players for {team_ref}")

    return team_name, players


def _team_from_xml_elm(team_elm, f7_root) -> Team:
    # This should not happen here
    team_name, team_players = _parse_team_players(
        f7_root, team_elm.attrib["TeamRef"]
    )

    team_id = team_elm.attrib["TeamRef"].lstrip("t")
    team = Team(
        team_id=str(team_id),
        name=team_name,
        ground=Ground.HOME
        if team_elm.attrib["Side"] == "Home"
        else Ground.AWAY,
    )
    team.players = [
        Player(
            player_id=player_elm.attrib["PlayerRef"].lstrip("p"),
            team=team,
            jersey_no=int(player_elm.attrib["ShirtNumber"]),
            first_name=team_players[player_elm.attrib["PlayerRef"]][
                "first_name"
            ],
            last_name=team_players[player_elm.attrib["PlayerRef"]][
                "last_name"
            ],
            position=Position(
                position_id=player_elm.attrib["Formation_Place"],
                name=player_elm.attrib["Position"],
                coordinates=None,
            ),
        )
        for player_elm in team_elm.find("PlayerLineUp").iterchildren(
            "MatchPlayer"
        )
    ]
    return team


def _get_event_qualifiers(raw_qualifiers: List) -> List[Qualifier]:
    qualifiers = []
    if EVENT_QUALIFIER_CORNER_KICK in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.CORNER_KICK))
    elif EVENT_QUALIFIER_FREE_KICK in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.FREE_KICK))
    elif EVENT_QUALIFIER_PENALTY in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.PENALTY))
    elif EVENT_QUALIFIER_THROW_IN in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.THROW_IN))
    elif EVENT_QUALIFIER_KICK_OFF in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.KICK_OFF))
    elif EVENT_QUALIFIER_GOAL_KICK in raw_qualifiers:
        qualifiers.append(SetPieceQualifier(value=SetPieceType.GOAL_KICK))

    return qualifiers


def _get_event_type_name(type_id: int) -> str:
    return event_type_names.get(type_id, "unknown")


class OptaSerializer(EventDataSerializer):
    @staticmethod
    def __validate_inputs(inputs: Dict[str, Readable]):
        if "f7_data" not in inputs:
            raise ValueError("Please specify a value for input 'f7_data'")
        if "f24_data" not in inputs:
            raise ValueError("Please specify a value for input 'f24_data'")

    def deserialize(
        self, inputs: Dict[str, Readable], options: Dict = None
    ) -> EventDataset:
        """
                Deserialize Opta event data into a `EventDataset`.

                Parameters
                ----------
                inputs : dict
                    input `f24_data` should point to a `Readable` object containing
                    the 'xml' formatted event data. input `f7_data` should point
                    to a `Readable` object containing the 'xml' formatted f7 data.
                options : dict
                    Options for deserialization of the Opta file. Possible options are
                    `event_types` (list of event types) to specify the event types that
                    should be returned. Valid types: "shot", "pass", "carry", "take_on" and
                    "generic". Generic is everything other than the first 4. Those events
                    are barely parsed. This type of event can be used to do the parsing
                    yourself.
                    Every event has a 'raw_event' attribute which contains the original
                    dictionary.
                Returns
                -------
                dataset : EventDataset
                Raises
                ------

                See Also
                --------

                Examples
                --------
                >>> serializer = OptaSerializer()
                >>> with open("123_f24.xml", "rb") as f24_data, \
                >>>      open("123_f7.xml", "rb") as f7_data:
                >>>
                >>>     dataset = serializer.deserialize(
                >>>         inputs={
                >>>             'f24_data': f24_data,
                >>>             'f7_data': f7_data
                >>>         },
                >>>         options={
                >>>             'event_types': ["pass", "take_on", "carry", "shot"]
                >>>         }
                >>>     )
                """
        self.__validate_inputs(inputs)
        if not options:
            options = {}

        with performance_logging("load data", logger=logger):
            f7_root = objectify.fromstring(inputs["f7_data"].read())
            f24_root = objectify.fromstring(inputs["f24_data"].read())

            wanted_event_types = [
                EventType[event_type.upper()]
                for event_type in options.get("event_types", [])
            ]

        with performance_logging("parse data", logger=logger):
            matchdata_path = objectify.ObjectPath(
                "SoccerFeed.SoccerDocument.MatchData"
            )
            team_elms = list(
                matchdata_path.find(f7_root).iterchildren("TeamData")
            )

            home_score = None
            away_score = None
            for team_elm in team_elms:
                if team_elm.attrib["Side"] == "Home":
                    home_score = team_elm.attrib["Score"]
                    home_team = _team_from_xml_elm(team_elm, f7_root)
                elif team_elm.attrib["Side"] == "Away":
                    away_score = team_elm.attrib["Score"]
                    away_team = _team_from_xml_elm(team_elm, f7_root)
                else:
                    raise Exception(f"Unknown side: {team_elm.attrib['Side']}")

            score = Score(home=home_score, away=away_score)
            teams = [home_team, away_team]

            if len(home_team.players) == 0 or len(away_team.players) == 0:
                raise Exception("LineUp incomplete")

            game_elm = f24_root.find("Game")
            periods = [
                Period(
                    id=1,
                    start_timestamp=None,
                    end_timestamp=None,
                ),
                Period(
                    id=2,
                    start_timestamp=None,
                    end_timestamp=None,
                ),
            ]
            possession_team = None
            events = []
            for event_elm in game_elm.iterchildren("Event"):
                event_id = event_elm.attrib["id"]
                type_id = int(event_elm.attrib["type_id"])
                timestamp = _parse_f24_datetime(event_elm.attrib["timestamp"])
                period_id = int(event_elm.attrib["period_id"])
                for period in periods:
                    if period.id == period_id:
                        break
                else:
                    logger.debug(
                        f"Skipping event {event_id} because period doesn't match {period_id}"
                    )
                    continue

                if type_id == EVENT_TYPE_START_PERIOD:
                    logger.debug(
                        f"Set start of period {period.id} to {timestamp}"
                    )
                    period.start_timestamp = timestamp
                elif type_id == EVENT_TYPE_END_PERIOD:
                    logger.debug(
                        f"Set end of period {period.id} to {timestamp}"
                    )
                    period.end_timestamp = timestamp
                else:
                    if not period.start_timestamp:
                        # not started yet
                        continue

                    if event_elm.attrib["team_id"] == home_team.team_id:
                        team = teams[0]
                    elif event_elm.attrib["team_id"] == away_team.team_id:
                        team = teams[1]
                    else:
                        raise Exception(
                            f"Unknown team_id {event_elm.attrib['team_id']}"
                        )

                    x = float(event_elm.attrib["x"])
                    y = float(event_elm.attrib["y"])
                    outcome = int(event_elm.attrib["outcome"])
                    raw_qualifiers = {
                        int(
                            qualifier_elm.attrib["qualifier_id"]
                        ): qualifier_elm.attrib.get("value")
                        for qualifier_elm in event_elm.iterchildren("Q")
                    }
                    player = None
                    if "player_id" in event_elm.attrib:
                        player = team.get_player_by_id(
                            event_elm.attrib["player_id"]
                        )

                    if type_id in BALL_OWNING_EVENTS:
                        possession_team = team

                    generic_event_kwargs = dict(
                        # from DataRecord
                        period=period,
                        timestamp=timestamp - period.start_timestamp,
                        ball_owning_team=possession_team,
                        ball_state=BallState.ALIVE,
                        # from Event
                        event_id=event_id,
                        team=team,
                        player=player,
                        coordinates=Point(x=x, y=y),
                        raw_event=event_elm,
                    )

                    if type_id == EVENT_TYPE_PASS:
                        pass_event_kwargs = _parse_pass(
                            raw_qualifiers, outcome
                        )
                        event = PassEvent.create(
                            **pass_event_kwargs,
                            **generic_event_kwargs,
                        )
                    elif type_id == EVENT_TYPE_OFFSIDE_PASS:
                        pass_event_kwargs = _parse_offside_pass(raw_qualifiers)
                        event = PassEvent.create(
                            **pass_event_kwargs,
                            **generic_event_kwargs,
                        )
                    elif type_id == EVENT_TYPE_TAKE_ON:
                        take_on_event_kwargs = _parse_take_on(outcome)
                        event = TakeOnEvent.create(
                            qualifiers=None,
                            **take_on_event_kwargs,
                            **generic_event_kwargs,
                        )
                    elif type_id in (
                        EVENT_TYPE_SHOT_MISS,
                        EVENT_TYPE_SHOT_POST,
                        EVENT_TYPE_SHOT_SAVED,
                        EVENT_TYPE_SHOT_GOAL,
                    ):
                        shot_event_kwargs = _parse_shot(
                            raw_qualifiers,
                            type_id,
                            coordinates=generic_event_kwargs["coordinates"],
                        )
                        kwargs = {}
                        kwargs.update(generic_event_kwargs)
                        kwargs.update(shot_event_kwargs)
                        event = ShotEvent.create(**kwargs)

                    elif type_id == EVENT_TYPE_RECOVERY:
                        event = RecoveryEvent.create(
                            result=None,
                            qualifiers=None,
                            **generic_event_kwargs,
                        )

                    elif type_id == EVENT_TYPE_FOUL_COMMITTED:
                        event = FoulCommittedEvent.create(
                            result=None,
                            qualifiers=None,
                            **generic_event_kwargs,
                        )

                    elif type_id in BALL_OUT_EVENTS:
                        generic_event_kwargs["ball_state"] = BallState.DEAD
                        event = BallOutEvent.create(
                            result=None,
                            qualifiers=None,
                            **generic_event_kwargs,
                        )

                    else:
                        event = GenericEvent.create(
                            **generic_event_kwargs,
                            result=None,
                            qualifiers=None,
                            event_name=_get_event_type_name(type_id),
                        )

                    if (
                        not wanted_event_types
                        or event.event_type in wanted_event_types
                    ):
                        events.append(event)

        metadata = Metadata(
            teams=teams,
            periods=periods,
            pitch_dimensions=PitchDimensions(
                x_dim=Dimension(0, 100), y_dim=Dimension(0, 100)
            ),
            score=score,
            frame_rate=None,
            orientation=Orientation.ACTION_EXECUTING_TEAM,
            flags=DatasetFlag.BALL_OWNING_TEAM,
            provider=Provider.OPTA,
        )

        return EventDataset(
            metadata=metadata,
            records=events,
        )

    def serialize(self, data_set: EventDataset) -> Tuple[str, str]:
        raise NotImplementedError



from dataclasses import dataclass
from typing import Set

from kloppy.domain import (
    Event,
    EventDataset,
    Player,
    SubstitutionEvent,
    PlayerOffEvent,
    PlayerOnEvent,
    CardEvent,
    CardType,
    Provider,
)
from ..builder import StateBuilder


@dataclass
class Lineup:
    players: Set[Player]


class LineupStateBuilder(StateBuilder):
    def initial_state(self, dataset: EventDataset) -> Lineup:
        return Lineup(
            players=(
                set(
                    player
                    for player in dataset.metadata.teams[0].players
                    if player.starting
                )
                | set(
                    player
                    for player in dataset.metadata.teams[1].players
                    if player.starting
                )
            )
        )

    def reduce_before(self, state: Lineup, event: Event) -> Lineup:
        return state

    def reduce_after(self, state: Lineup, event: Event) -> Lineup:
        if isinstance(event, SubstitutionEvent):
            state = Lineup(
                players=state.players - {event.player}
                | {event.replacement_player}
            )
        elif isinstance(event, PlayerOffEvent):
            state = Lineup(players=state.players - {event.player})
        elif isinstance(event, PlayerOnEvent):
            state = Lineup(players=state.players | {event.player})
        elif isinstance(event, CardEvent):
            if event.card_type in (CardType.SECOND_YELLOW, CardType.RED):
                state = Lineup(players=state.players - {event.player})
        return state



from lxml import objectify
import warnings

from kloppy.domain import (
    Period,
    PitchDimensions,
    Dimension,
    Team,
    Score,
    Ground,
    DatasetFlag,
    AttackingDirection,
    Orientation,
    Position,
    Point,
    Provider,
)
from kloppy.utils import Readable

from .models import *


def noop(x):
    return x


def _load_provider_parameters(parent_elm, value_mapper=None) -> Dict:
    if parent_elm is None:
        return {}

    if not value_mapper:
        value_mapper = noop

    return {
        str(param.find("Name")): value_mapper(param.find("Value"))
        for param in parent_elm.iterchildren(tag="ProviderParameter")
        if param.find("Value") != ""
    }


def _load_periods(global_config_elm, frame_rate: int) -> List[Period]:
    provider_params = _load_provider_parameters(
        global_config_elm.find("ProviderGlobalParameters"), value_mapper=int
    )

    period_names = [
        "first_half",
        "second_half",
        "first_extra_half",
        "second_extra_half",
    ]

    periods = []

    for idx, period_name in enumerate(period_names):
        start_key = f"{period_name}_start"
        end_key = f"{period_name}_end"
        if start_key in provider_params:
            periods.append(
                Period(
                    id=idx + 1,
                    start_timestamp=float(provider_params[start_key])
                    / frame_rate,
                    end_timestamp=float(provider_params[end_key]) / frame_rate,
                )
            )
        else:
            # done
            break

    return periods


def _load_players(players_elm, team: Team) -> List[Player]:
    return [
        Player(
            team=team,
            jersey_no=int(player_elm.find("ShirtNumber")),
            player_id=player_elm.attrib["id"],
            name=str(player_elm.find("Name")),
            position=_load_position_data(
                player_elm.find("ProviderPlayerParameters")
            ),
            attributes=_load_provider_parameters(
                player_elm.find("ProviderPlayerParameters")
            ),
        )
        for player_elm in players_elm.iterchildren(tag="Player")
        if player_elm.attrib["teamId"] == team.team_id
    ]


def _load_position_data(parent_elm) -> Position:
    # TODO: _load_provider_parameters is called twice to set position data
    # and then again to set the attributes. Also, data in position should not
    # be duplicated in attributes either.
    player_provider_parameters = _load_provider_parameters(parent_elm)
    if "position_index" not in player_provider_parameters:
        return None

    return Position(
        position_id=player_provider_parameters["position_index"],
        name=player_provider_parameters["position_type"],
        coordinates=Point(
            player_provider_parameters["position_x"],
            player_provider_parameters["position_y"],
        ),
    )


def _load_data_format_specifications(
    data_format_specifications_elm,
) -> List[DataFormatSpecification]:
    return [
        DataFormatSpecification.from_xml_element(data_format_specification_elm)
        for data_format_specification_elm in data_format_specifications_elm.iterchildren(
            tag="DataFormatSpecification"
        )
    ]


def _load_sensors(sensors_elm) -> List[Sensor]:
    return [
        Sensor.from_xml_element(sensor_elm)
        for sensor_elm in sensors_elm.iterchildren(tag="Sensor")
    ]


def _load_pitch_dimensions(
    metadata_elm, sensors: List[Sensor]
) -> Union[None, PitchDimensions]:

    normalized = False
    for sensor in sensors:
        if sensor.sensor_id == "position":
            if sensor.channels[0].unit == "normalized":
                normalized = True
                break

    field_size_path = objectify.ObjectPath("Metadata.Sessions.Session[0]")
    field_size_elm = field_size_path.find(metadata_elm).find("FieldSize")

    if field_size_elm is not None and normalized:
        return PitchDimensions(
            x_dim=Dimension(0, 1),
            y_dim=Dimension(0, 1),
            x_per_meter=1 / int(field_size_elm.find("Width")),
            y_per_meter=1 / int(field_size_elm.find("Height")),
        )
    else:
        return None


def _parse_provider(provider_name: Union[str, None]) -> Provider:
    if provider_name is not None:
        if provider_name == "Metrica Sports":
            return Provider.METRICA
        else:
            warnings.warn(
                "The Provider is not known to Kloppy.",
                Warning,
            )
    else:
        return None


def _load_provider(metadata_elm, provider: Provider = None) -> Provider:
    provider_path = objectify.ObjectPath("Metadata.GlobalConfig.ProviderName")
    provider_name = provider_path.find(metadata_elm)
    provider_from_file = _parse_provider(provider_name)
    if provider:
        if provider_from_file and provider_from_file != provider:
            warnings.warn(
                f"Given provider name is different to the name of the Provider read from the XML-file",
                Warning,
            )
    else:
        provider = provider_from_file
    return provider


def load_metadata(
    metadata_file: Readable, provider: Provider = None
) -> EPTSMetadata:
    root = objectify.fromstring(metadata_file.read())
    metadata = root.find("Metadata")

    provider = _load_provider(metadata, provider)

    score_path = objectify.ObjectPath(
        "Metadata.Sessions.Session[0].MatchParameters.Score"
    )
    score_elm = score_path.find(metadata)
    score = Score(
        home=score_elm.LocalTeamScore, away=score_elm.VisitingTeamScore
    )

    _team_map = {
        Ground.HOME: score_elm.attrib["idLocalTeam"],
        Ground.AWAY: score_elm.attrib["idVisitingTeam"],
    }

    _team_name_map = {
        team_elm.attrib["id"]: str(team_elm.find("Name"))
        for team_elm in metadata.find("Teams").iterchildren(tag="Team")
    }

    teams_metadata = {}
    for ground, team_id in _team_map.items():
        team = Team(
            team_id=team_id, name=_team_name_map[team_id], ground=ground
        )
        team.players = _load_players(metadata.find("Players"), team)
        teams_metadata.update({ground: team})

    data_format_specifications = _load_data_format_specifications(
        root.find("DataFormatSpecifications")
    )

    device_path = objectify.ObjectPath("Metadata.Devices.Device[0].Sensors")
    sensors = _load_sensors(device_path.find(metadata))

    _channel_map = {
        channel.channel_id: channel
        for sensor in sensors
        for channel in sensor.channels
    }

    _all_players = [
        player
        for key, value in teams_metadata.items()
        for player in value.players
    ]

    _player_map = {player.player_id: player for player in _all_players}

    player_channels = [
        PlayerChannel(
            player_channel_id=player_channel_elm.attrib["id"],
            player=_player_map[player_channel_elm.attrib["playerId"]],
            channel=_channel_map[player_channel_elm.attrib["channelId"]],
        )
        for player_channel_elm in metadata.find("PlayerChannels").iterchildren(
            tag="PlayerChannel"
        )
    ]

    frame_rate = int(metadata.find("GlobalConfig").find("FrameRate"))
    pitch_dimensions = _load_pitch_dimensions(metadata, sensors)
    periods = _load_periods(metadata.find("GlobalConfig"), frame_rate)

    if periods:
        start_attacking_direction = periods[0].attacking_direction
    else:
        start_attacking_direction = None

    orientation = (
        (
            Orientation.FIXED_HOME_AWAY
            if start_attacking_direction == AttackingDirection.HOME_AWAY
            else Orientation.FIXED_AWAY_HOME
        )
        if start_attacking_direction != AttackingDirection.NOT_SET
        else None
    )

    metadata.orientation = orientation

    return EPTSMetadata(
        teams=list(teams_metadata.values()),
        periods=periods,
        pitch_dimensions=pitch_dimensions,
        data_format_specifications=data_format_specifications,
        player_channels=player_channels,
        frame_rate=frame_rate,
        sensors=sensors,
        score=score,
        orientation=None,
        provider=provider,
        flags=~(DatasetFlag.BALL_STATE | DatasetFlag.BALL_OWNING_TEAM),
    )



from abc import ABC, abstractmethod
from typing import Tuple, Dict

from kloppy.utils import Readable
from kloppy.domain import EventDataset


class EventDataSerializer(ABC):
    @abstractmethod
    def deserialize(
        self, inputs: Dict[str, Readable], options: Dict = None
    ) -> EventDataset:
        raise NotImplementedError

    @abstractmethod
    def serialize(self, dataset: EventDataset) -> Tuple[str, str]:
        raise NotImplementedError



import os

from kloppy import TRACABSerializer
from kloppy.domain import (
    Period,
    AttackingDirection,
    Orientation,
    Provider,
    Point,
    Point3D,
    BallState,
    Team,
    Ground,
)
from kloppy.domain.models.common import DatasetType


class TestTracabTracking:
    def test_correct_deserialization(self):
        base_dir = os.path.dirname(__file__)

        serializer = TRACABSerializer()

        with open(f"{base_dir}/files/tracab_meta.xml", "rb") as metadata, open(
            f"{base_dir}/files/tracab_raw.dat", "rb"
        ) as raw_data:

            dataset = serializer.deserialize(
                inputs={"metadata": metadata, "raw_data": raw_data},
                options={"only_alive": False},
            )

        assert dataset.metadata.provider == Provider.TRACAB
        assert dataset.dataset_type == DatasetType.TRACKING
        assert len(dataset.records) == 6
        assert len(dataset.metadata.periods) == 2
        assert dataset.metadata.orientation == Orientation.FIXED_HOME_AWAY
        assert dataset.metadata.periods[0] == Period(
            id=1,
            start_timestamp=4.0,
            end_timestamp=4.08,
            attacking_direction=AttackingDirection.HOME_AWAY,
        )

        assert dataset.metadata.periods[1] == Period(
            id=2,
            start_timestamp=8.0,
            end_timestamp=8.08,
            attacking_direction=AttackingDirection.AWAY_HOME,
        )

        player_home_19 = dataset.metadata.teams[0].get_player_by_jersey_number(
            "19"
        )
        assert dataset.records[0].players_coordinates[player_home_19] == Point(
            x=-1234.0, y=-294.0
        )

        player_away_19 = dataset.metadata.teams[1].get_player_by_jersey_number(
            "19"
        )
        assert dataset.records[0].players_coordinates[player_away_19] == Point(
            x=8889, y=-666
        )
        assert dataset.records[0].ball_coordinates == Point3D(x=-27, y=25, z=0)
        assert dataset.records[0].ball_state == BallState.ALIVE
        assert dataset.records[0].ball_owning_team == Team(
            team_id="home", name="home", ground=Ground.HOME
        )

        assert dataset.records[1].ball_owning_team == Team(
            team_id="away", name="away", ground=Ground.AWAY
        )

        assert dataset.records[2].ball_state == BallState.DEAD

        # make sure player data is only in the frame when the player is at the pitch
        assert "away_1337" not in [
            player.player_id
            for player in dataset.records[0].players_coordinates.keys()
        ]
        assert "away_1337" in [
            player.player_id
            for player in dataset.records[3].players_coordinates.keys()
        ]



import os

from kloppy import (
    MetricaTrackingSerializer,
    MetricaEventsJsonSerializer,
    load_sportec_event_data,
)
from kloppy.domain import (
    Period,
    Provider,
    AttackingDirection,
    Orientation,
    Point,
    EventType,
)
from kloppy.domain.models.common import DatasetType


class TestSportecEvent:
    def test_correct_deserialization(self):
        base_dir = os.path.dirname(__file__)

        dataset = load_sportec_event_data(
            f"{base_dir}/files/sportec_events.xml",
            f"{base_dir}/files/sportec_meta.xml",
        )

        assert dataset.metadata.provider == Provider.SPORTEC
        assert dataset.dataset_type == DatasetType.EVENT
        assert len(dataset.metadata.periods) == 2

        # raw_event must be flattened dict
        assert isinstance(dataset.events[0].raw_event, dict)

        assert len(dataset.events) == 28
        assert dataset.metadata.orientation == Orientation.FIXED_HOME_AWAY
        assert dataset.metadata.periods[0] == Period(
            id=1,
            start_timestamp=1591381800.21,
            end_timestamp=1591384584.0,
            attacking_direction=AttackingDirection.HOME_AWAY,
        )
        assert dataset.metadata.periods[1] == Period(
            id=2,
            start_timestamp=1591385607.01,
            end_timestamp=1591388598.0,
            attacking_direction=AttackingDirection.AWAY_HOME,
        )

        player = dataset.metadata.teams[0].players[0]
        assert player.player_id == "DFL-OBJ-00001D"
        assert player.jersey_no == 1
        assert str(player) == "A. Schwolow"
        assert player.position.position_id is None
        assert player.position.name == "TW"



from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from typing import Callable, Tuple, Dict, List, Iterator

from kloppy.domain import (
    EventDataset,
    PassEvent,
    ShotEvent,
    CarryEvent,
    TakeOnEvent,
    Event,
)
from .regexp import *
from .regexp import _make_match, _TrailItem
from .regexp.regexp import _Match


class WithCaptureMatcher(Matcher):
    def __init__(self, matcher: Callable[[Tok, Dict[str, List[Tok]]], bool]):
        self.matcher = matcher

    def _add_captures(self, captures: Dict[str, List[Tok]], match: _Match):
        for name, capture in match.children.items():
            captures[name] = capture[0].trail
            self._add_captures(captures, capture[0])

    def match(
        self, token: Tok, trail: Tuple[_TrailItem[Out], ...]
    ) -> Iterator[Out]:
        match = _make_match(trail)
        captures = {}
        self._add_captures(captures, match)
        if self.matcher(token, captures):
            yield token


def match_generic(event_cls, capture=None, **kwargs):
    def _matcher_fn(event: Event, captures: Dict[str, List[Event]]) -> bool:
        if not isinstance(event, event_cls):
            return False

        # TODO: v[0] points to first record
        captures = {k: v[0] for k, v in captures.items()}
        for attr_name, attr_value in kwargs.items():
            if callable(attr_value):
                attr_real_value = getattr(event, attr_name)
                result = attr_value(attr_name, attr_real_value, captures)
            else:
                if attr_name == "success":
                    result = event.result and event.result.is_success
                else:
                    result = getattr(event, attr_name) == attr_value
            if not result:
                return False
        return True

    _matcher = Final(WithCaptureMatcher(matcher=_matcher_fn))

    if capture:
        return _matcher[capture]
    else:
        return _matcher


match_pass = partial(match_generic, PassEvent)
match_shot = partial(match_generic, ShotEvent)
match_carry = partial(match_generic, CarryEvent)
match_take_on = partial(match_generic, TakeOnEvent)
match_any = partial(match_generic, Event)


def same_as(capture: str):
    capture_name, attribute_name = capture.split(".")

    def fn(attr_name, value, captures):
        return value == getattr(captures[capture_name], attribute_name)

    return fn


def not_same_as(capture: str):
    capture_name, capture_attribute_name = capture.split(".")

    def fn(attr_name, value, captures):
        return value != getattr(captures[capture_name], capture_attribute_name)

    return fn


def group(node, capture=None):
    if capture:
        return node[capture]
    return node


def function(fn):
    def wrapper(attr_name, value, captures):
        capture_values = {
            f"{capture_name}_{attr_name}": getattr(capture_value, attr_name)
            for capture_name, capture_value in captures.items()
            if capture_value
        }
        return fn(value, **capture_values)

    return wrapper


@dataclass
class Match:
    events: List[Event]
    captures: Dict[str, List[Event]]


def search(dataset: EventDataset, pattern: Node[Tok, Out]):
    events = dataset.events
    re = RegExp.from_ast(pattern)

    results = []
    events_per_period = defaultdict(list)
    for event in events:
        events_per_period[event.period.id].append(event)

    for period, events_ in sorted(events_per_period.items()):
        # Search per period. Patterns should never match over periods
        results.extend(_search(events_, re))
    return results


def _search(events: List[Event], re: RegExp[Tok, Out]):
    i = 0
    results = []
    for i in range(len(events)):
        matches = re.match(events[i:], consume_all=False)
        if matches:
            results.append(
                Match(
                    events=matches[0].trail,
                    # TODO: check trail[0] because this points to the first event in the capture and not
                    #       all of them
                    captures={
                        capture_name: capture_value[0].trail[0]
                        for capture_name, capture_value in matches[
                            0
                        ].children.items()
                    },
                )
            )

    return results


@dataclass
class Query:
    event_types: List[str]
    pattern: Node[Tok, Out]


__all__ = [
    "search",
    "match_pass",
    "match_carry",
    "match_take_on",
    "match_shot",
    "match_any",
    "same_as",
    "not_same_as",
    "function",
    "Query",
]



import os

from kloppy import StatsBombSerializer
from kloppy.domain import (
    AttackingDirection,
    Period,
    Orientation,
    Provider,
    EventType,
    BodyPartQualifier,
    BodyPart,
)
from kloppy.domain.models.common import DatasetType


class TestStatsbomb:
    def _load_dataset(self, options=None):
        base_dir = os.path.dirname(__file__)

        serializer = StatsBombSerializer()

        with open(
            f"{base_dir}/files/statsbomb_lineup.json", "rb"
        ) as lineup_data, open(
            f"{base_dir}/files/statsbomb_event.json", "rb"
        ) as event_data:
            dataset = serializer.deserialize(
                inputs={"lineup_data": lineup_data, "event_data": event_data},
                options=options,
            )
            return dataset

    def test_correct_deserialization(self):
        """
        This test uses data from the StatsBomb open data project.
        """
        dataset = self._load_dataset()

        assert dataset.metadata.provider == Provider.STATSBOMB
        assert dataset.dataset_type == DatasetType.EVENT
        assert len(dataset.events) == 4022
        assert len(dataset.metadata.periods) == 2
        assert (
            dataset.metadata.orientation == Orientation.ACTION_EXECUTING_TEAM
        )
        assert dataset.metadata.teams[0].name == "Barcelona"
        assert dataset.metadata.teams[1].name == "Deportivo Alavés"

        player = dataset.metadata.teams[0].get_player_by_id("5503")
        assert player.player_id == "5503"
        assert player.jersey_no == 10
        assert str(player) == "Lionel Andrés Messi Cuccittini"
        assert player.position is None  # not set
        assert player.starting

        sub_player = dataset.metadata.teams[0].get_player_by_id("3501")
        assert str(sub_player) == "Philippe Coutinho Correia"
        assert not sub_player.starting

        assert dataset.metadata.periods[0] == Period(
            id=1,
            start_timestamp=0.0,
            end_timestamp=2705.267,
            attacking_direction=AttackingDirection.NOT_SET,
        )
        assert dataset.metadata.periods[1] == Period(
            id=2,
            start_timestamp=2705.268,
            end_timestamp=5557.321,
            attacking_direction=AttackingDirection.NOT_SET,
        )

        assert (
            dataset.events[791].get_qualifier_value(BodyPartQualifier)
            == BodyPart.HEAD
        )

        assert (
            dataset.events[2231].get_qualifier_value(BodyPartQualifier)
            == BodyPart.RIGHT_FOOT
        )

        assert (
            dataset.events[195].get_qualifier_value(BodyPartQualifier) is None
        )

    def test_substitution(self):
        """
        Test substitution events
        """

        dataset = self._load_dataset({"event_types": ["substitution"]})

        assert len(dataset.events) == 6

        subs = [
            (6374, 3501),
            (6839, 6935),
            (6581, 6566),
            (6613, 6624),
            (5477, 11392),
            (5203, 8206),
        ]

        for event_idx, (player_id, replacement_player_id) in enumerate(subs):
            event = dataset.events[event_idx]
            assert event.player == event.team.get_player_by_id(player_id)
            assert event.replacement_player == event.team.get_player_by_id(
                replacement_player_id
            )



import setuptools
import builtins
from distutils.core import setup


builtins.__KLOPPY_SETUP__ = True
import kloppy


def setup_package():
    with open("README.md", "r") as f:
        readme = f.read()

    setup(
        name="kloppy",
        version=kloppy.__version__,
        author="Koen Vossen",
        author_email="info@koenvossen.nl",
        url="https://kloppy.pysport.org/",
        packages=setuptools.find_packages(exclude=["tests"]),
        license="BSD",
        description="Standardizing soccer tracking- and event data",
        long_description=readme,
        long_description_content_type="text/markdown",
        classifiers=[
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "License :: OSI Approved",
            "Topic :: Scientific/Engineering",
        ],
        entry_points={
            "console_scripts": ["kloppy-query = kloppy.cmdline:run_query"]
        },
        install_requires=[
            "lxml>=4.5.0",
            "requests>=2.0.0",
            "networkx>=2.4",
            "pytz>=2020.1",
            'dataclasses;python_version<"3.7"',
        ],
        extras_require={
            "test": ["pytest", "pandas>=1.0.0"],
            "development": ["pre-commit"],
        },
    )


if __name__ == "__main__":
    setup_package()

    del builtins.__KLOPPY_SETUP__



import logging
from typing import Tuple, Dict, Union

from lxml import objectify, etree


from kloppy.domain import (
    CodeDataset,
    Code,
    Period,
    Metadata,
    Provider,
    DatasetFlag,
    Score,
    PitchDimensions,
    Orientation,
)
from kloppy.utils import Readable

from .base import CodeDataSerializer

logger = logging.getLogger(__name__)


def parse_value(s: str) -> Union[str, float, int]:
    try:
        func = float if "." in s else int
        return func(s)
    except ValueError:
        return s


class XMLCodeSerializer(CodeDataSerializer):
    def deserialize(
        self, inputs: Dict[str, Readable], options: Dict = None
    ) -> CodeDataset:
        all_instances = objectify.fromstring(inputs["xml_file"].read())

        codes = []
        period = Period(id=1, start_timestamp=0, end_timestamp=0)
        for instance in all_instances.ALL_INSTANCES.iterchildren():
            end_timestamp = float(instance.end)

            code = Code(
                period=period,
                code_id=str(instance.ID),
                code=str(instance.code),
                timestamp=float(instance.start),
                end_timestamp=end_timestamp,
                labels={
                    str(label.find("group")): parse_value(
                        str(label.find("text"))
                    )
                    for label in instance.iterchildren("label")
                },
                ball_state=None,
                ball_owning_team=None,
            )
            period.end_timestamp = end_timestamp
            codes.append(code)

        return CodeDataset(
            metadata=Metadata(
                teams=[],
                periods=[period],
                pitch_dimensions=PitchDimensions.default(),
                score=Score(0, 0),
                frame_rate=0.0,
                orientation=Orientation.NOT_SET,
                flags=~(DatasetFlag.BALL_OWNING_TEAM | DatasetFlag.BALL_STATE),
                provider=Provider.OTHER,
            ),
            records=codes,
        )

    def serialize(self, dataset: CodeDataset) -> bytes:
        root = etree.Element("file")
        all_instances = etree.SubElement(root, "ALL_INSTANCES")
        for i, code in enumerate(dataset.codes):
            relative_period_start = 0
            for period in dataset.metadata.periods:
                if period == code.period:
                    break
                else:
                    relative_period_start += period.duration

            instance = etree.SubElement(all_instances, "instance")
            id_ = etree.SubElement(instance, "ID")
            id_.text = code.code_id or str(i + 1)

            start = etree.SubElement(instance, "start")
            start.text = str(relative_period_start + code.start_timestamp)

            end = etree.SubElement(instance, "end")
            end.text = str(relative_period_start + code.end_timestamp)

            code_ = etree.SubElement(instance, "code")
            code_.text = code.code

            for group, text in code.labels.items():
                label = etree.SubElement(instance, "label")
                group_ = etree.SubElement(label, "group")
                group_.text = group

                text_ = etree.SubElement(label, "text")
                text_.text = str(text)

        return etree.tostring(
            root,
            pretty_print=True,
            xml_declaration=True,
            encoding="utf-8",  # This might not work with some tools because they expected 'ascii'.
            method="xml",
        )



import argparse
import logging
import sys
import textwrap
from collections import Counter

from kloppy import (
    load_opta_event_data,
    load_statsbomb_event_data,
    load_wyscout_event_data,
    write_xml_code_data,
    event_pattern_matching as pm,
)
from kloppy.domain import CodeDataset, Code
from kloppy.utils import performance_logging

sys.path.append(".")


def _format_time(seconds: float) -> str:
    minutes, seconds = divmod(seconds, 60)
    return f"{minutes:02.0f}:{seconds:02.0f}"


def print_match(id_: int, match, success: bool, label):
    print(f"Match {id_}: {label} {'SUCCESS' if success else 'no-success'}")
    for event in match.events:
        time = _format_time(event.timestamp)
        print(
            f"{event.event_id} {event.event_type} {str(event.result).ljust(10)} / P{event.period.id} {time} / {event.team} {str(event.player.jersey_no).rjust(2)} / {event.coordinates.x}x{event.coordinates.y}"
        )
    print("")


def load_query(query_file: str) -> pm.Query:
    locals_dict = {}
    with open(query_file, "rb") as fp:
        exec(fp.read(), {}, locals_dict)

    if "query" not in locals_dict:
        raise Exception("File does not contain query")
    return locals_dict["query"]


def run_query(argv=sys.argv[1:]):
    parser = argparse.ArgumentParser(description="Run query on event data")
    parser.add_argument(
        "--input-statsbomb",
        help="StatsBomb event input files (events.json,lineup.json)",
    )
    parser.add_argument(
        "--input-opta", help="Opta event input files (f24.xml,f7.xml)"
    )
    parser.add_argument("--input-wyscout", help="Wyscout event input file")
    parser.add_argument("--output-xml", help="Output file")
    parser.add_argument(
        "--with-success",
        default=True,
        help="Input existence of success capture in output",
    )
    parser.add_argument(
        "--prepend-time", default=7, help="Seconds to prepend to match"
    )
    parser.add_argument(
        "--append-time", default=5, help="Seconds to append to match"
    )
    parser.add_argument(
        "--query-file", help="File containing the query", required=True
    )
    parser.add_argument(
        "--stats",
        default="none",
        help="Show matches stats",
        choices=["text", "json", "none"],
    )
    parser.add_argument(
        "--show-events",
        default=False,
        help="Show events for each match",
        action="store_true",
    )
    parser.add_argument(
        "--only-success",
        default=False,
        help="Only show/output success cases",
        action="store_true",
    )

    logger = logging.getLogger("run_query")
    logging.basicConfig(
        stream=sys.stderr,
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    opts = parser.parse_args(argv)

    query = load_query(opts.query_file)

    dataset = None
    if opts.input_statsbomb:
        with performance_logging("load dataset", logger=logger):
            events_filename, lineup_filename = opts.input_statsbomb.split(",")
            dataset = load_statsbomb_event_data(
                events_filename.strip(),
                lineup_filename.strip(),
                options={"event_types": query.event_types},
            )
    if opts.input_opta:
        with performance_logging("load dataset", logger=logger):
            f24_filename, f7_filename = opts.input_opta.split(",")
            dataset = load_opta_event_data(
                f24_filename.strip(),
                f7_filename.strip(),
                options={"event_types": query.event_types},
            )
    if opts.input_wyscout:
        with performance_logging("load dataset", logger=logger):
            events_filename = opts.input_wyscout
            dataset = load_wyscout_event_data(
                events_filename,
                options={"event_types": query.event_types},
            )

    if not dataset:
        raise Exception("You have to specify a dataset.")

    with performance_logging("searching", logger=logger):
        matches = pm.search(dataset, query.pattern)

    # Construct new code dataset with same properties (eg periods)
    # as original event dataset.
    # Records will be added later below
    code_dataset = CodeDataset(metadata=dataset.metadata, records=[])

    counter = Counter()
    for i, match in enumerate(matches):
        team = match.events[0].team
        success = "success" in match.captures

        counter.update(
            {
                f"{team.ground}_total": 1,
                f"{team.ground}_success": 1 if success else 0,
            }
        )

        should_process = not opts.only_success or success
        if opts.show_events and should_process:
            print_match(i, match, success, str(team))

        if opts.output_xml and should_process:
            code_ = str(team)
            if opts.with_success and success:
                code_ += " success"

            code = Code(
                period=match.events[0].period,
                code_id=str(i),
                code=code_,
                timestamp=match.events[0].timestamp - opts.prepend_time,
                end_timestamp=match.events[-1].timestamp + opts.append_time,
                # TODO: refactor those two out
                ball_state=None,
                ball_owning_team=None,
            )
            code_dataset.records.append(code)

    if opts.output_xml:
        write_xml_code_data(code_dataset, opts.output_xml)
        logger.info(f"Wrote {len(code_dataset.codes)} video fragments to file")

    if opts.stats == "text":
        text_stats = """\
        Home:
          total count: {home_total}
            success: {home_success} ({home_success_rate:.0f}%)
            no success: {home_failure} ({home_failure_rate:.0f}%)

        Away:
          total count: {away_total}
            success: {away_success} ({away_success_rate:.0f}%)
            no success: {away_failure} ({away_failure_rate:.0f}%)
        """.format(
            home_total=counter["home_total"],
            home_success=counter["home_success"],
            home_success_rate=(
                counter["home_success"] / counter["home_total"] * 100
            ),
            home_failure=counter["home_total"] - counter["home_success"],
            home_failure_rate=(
                (counter["home_total"] - counter["home_success"])
                / counter["home_total"]
                * 100
            ),
            away_total=counter["away_total"],
            away_success=counter["away_success"],
            away_success_rate=(
                counter["away_success"] / counter["away_total"] * 100
            ),
            away_failure=counter["away_total"] - counter["away_success"],
            away_failure_rate=(
                (counter["away_total"] - counter["away_success"])
                / counter["away_total"]
                * 100
            ),
        )
        print(textwrap.dedent(text_stats))
    elif opts.stats == "json":
        import json

        print(json.dumps(counter, indent=4))



import logging
from typing import Tuple, Dict

from kloppy.domain import (
    TrackingDataset,
    DatasetFlag,
    AttackingDirection,
    Frame,
    Point,
    Point3D,
    Team,
    Orientation,
)
from kloppy.utils import Readable, performance_logging

from .metadata import load_metadata, EPTSMetadata
from .reader import read_raw_data

from .. import TrackingDataSerializer

logger = logging.getLogger(__name__)


class EPTSSerializer(TrackingDataSerializer):
    @staticmethod
    def __validate_inputs(inputs: Dict[str, Readable]):
        if "metadata" not in inputs:
            raise ValueError("Please specify a value for 'metadata'")
        if "raw_data" not in inputs:
            raise ValueError("Please specify a value for 'raw_data'")

    @staticmethod
    def _frame_from_row(row: dict, metadata: EPTSMetadata) -> Frame:
        timestamp = row["timestamp"]
        if metadata.periods and row["period_id"]:
            # might want to search for it instead
            period = metadata.periods[row["period_id"] - 1]
        else:
            period = None

        players_coordinates = {}
        for team in metadata.teams:
            for player in team.players:
                if f"player_{player.player_id}_x" in row:
                    players_coordinates[player] = Point(
                        x=row[f"player_{player.player_id}_x"],
                        y=row[f"player_{player.player_id}_y"],
                    )

        return Frame(
            frame_id=row["frame_id"],
            timestamp=timestamp,
            ball_owning_team=None,
            ball_state=None,
            period=period,
            players_coordinates=players_coordinates,
            ball_coordinates=Point3D(
                x=row["ball_x"], y=row["ball_y"], z=row["ball_z"]
            ),
        )

    def deserialize(
        self, inputs: Dict[str, Readable], options: Dict = None
    ) -> TrackingDataset:
        """
        Deserialize EPTS tracking data into a `TrackingDataset`.

        Parameters
        ----------
        inputs : dict
            input `raw_data` should point to a `Readable` object containing
            the 'csv' formatted raw data. input `metadata` should point to
            the xml metadata data.
        options : dict
            Options for deserialization of the EPTS file. Possible options are
            `sample_rate` (float between 0 and 1) to specify the amount of
            frames that should be loaded, `limit` to specify the max number of
            frames that will be returned.
        Returns
        -------
        dataset : TrackingDataset
        Raises
        ------
        -

        See Also
        --------

        Examples
        --------
        >>> serializer = EPTSSerializer()
        >>> with open("metadata.xml", "rb") as meta, \
        >>>      open("raw.dat", "rb") as raw:
        >>>     dataset = serializer.deserialize(
        >>>         inputs={
        >>>             'metadata': meta,
        >>>             'raw_data': raw
        >>>         },
        >>>         options={
        >>>             'sample_rate': 1/12
        >>>         }
        >>>     )
        """
        self.__validate_inputs(inputs)

        if not options:
            options = {}

        sample_rate = float(options.get("sample_rate", 1.0))
        limit = int(options.get("limit", 0))

        with performance_logging("Loading metadata", logger=logger):
            metadata = load_metadata(inputs["metadata"])

        with performance_logging("Loading data", logger=logger):
            # assume they are sorted
            frames = [
                self._frame_from_row(row, metadata)
                for row in read_raw_data(
                    raw_data=inputs["raw_data"],
                    metadata=metadata,
                    sensor_ids=[
                        "position"
                    ],  # we don't care about other sensors
                    sample_rate=sample_rate,
                    limit=limit,
                )
            ]

        return TrackingDataset(records=frames, metadata=metadata)

    def serialize(self, dataset: TrackingDataset) -> Tuple[str, str]:
        raise NotImplementedError



from dataclasses import asdict, fields, replace
from typing import TypeVar

from kloppy.domain import (
    AttackingDirection,
    Dataset,
    DatasetFlag,
    EventDataset,
    Frame,
    Metadata,
    Orientation,
    PitchDimensions,
    Point,
    Team,
    TrackingDataset,
)
from kloppy.domain.models.event import Event


class Transformer:
    def __init__(
        self,
        from_pitch_dimensions: PitchDimensions,
        from_orientation: Orientation,
        to_pitch_dimensions: PitchDimensions,
        to_orientation: Orientation,
    ):
        self._from_pitch_dimensions = from_pitch_dimensions
        self._from_orientation = from_orientation
        self._to_pitch_dimensions = to_pitch_dimensions
        self._to_orientation = to_orientation

    def transform_point(self, point: Point, flip: bool) -> Point:
        # 1. always apply changes from coordinate system
        # 2. flip coordinates depending on orientation
        if point is None:
            return None
        x_base = self._from_pitch_dimensions.x_dim.to_base(point.x)
        y_base = self._from_pitch_dimensions.y_dim.to_base(point.y)

        if flip:
            x_base = 1 - x_base
            y_base = 1 - y_base

        return Point(
            x=self._to_pitch_dimensions.x_dim.from_base(x_base),
            y=self._to_pitch_dimensions.y_dim.from_base(y_base),
        )

    def __needs_flip(
        self,
        ball_owning_team: Team,
        attacking_direction: AttackingDirection,
        action_executing_team: Team = None,
    ) -> bool:
        if self._from_orientation == self._to_orientation:
            flip = False
        else:
            orientation_factor_from = (
                self._from_orientation.get_orientation_factor(
                    ball_owning_team=ball_owning_team,
                    attacking_direction=attacking_direction,
                    action_executing_team=action_executing_team,
                )
            )
            orientation_factor_to = (
                self._to_orientation.get_orientation_factor(
                    ball_owning_team=ball_owning_team,
                    attacking_direction=attacking_direction,
                    action_executing_team=action_executing_team,
                )
            )
            flip = orientation_factor_from != orientation_factor_to
        return flip

    def transform_frame(self, frame: Frame) -> Frame:
        flip = self.__needs_flip(
            ball_owning_team=frame.ball_owning_team,
            attacking_direction=frame.period.attacking_direction,
        )

        return Frame(
            # doesn't change
            timestamp=frame.timestamp,
            frame_id=frame.frame_id,
            ball_owning_team=frame.ball_owning_team,
            ball_state=frame.ball_state,
            period=frame.period,
            # changes
            ball_coordinates=self.transform_point(
                frame.ball_coordinates, flip
            ),
            players_coordinates={
                key: self.transform_point(point, flip)
                for key, point in frame.players_coordinates.items()
            },
        )

    def transform_event(self, event: Event) -> Event:
        flip = self.__needs_flip(
            ball_owning_team=event.ball_owning_team,
            attacking_direction=event.period.attacking_direction,
            action_executing_team=event.team,
        )

        position_changes = {
            field.name: self.transform_point(getattr(event, field.name), flip)
            for field in fields(event)
            if field.name.endswith("coordinates")
            and getattr(event, field.name)
        }

        return replace(event, **position_changes)

    DatasetT = TypeVar("DatasetT")

    @classmethod
    def transform_dataset(
        cls,
        dataset: DatasetT,
        to_pitch_dimensions: PitchDimensions = None,
        to_orientation: Orientation = None,
    ) -> DatasetT:
        if not to_pitch_dimensions and not to_orientation:
            return dataset
        elif not to_orientation:
            to_orientation = dataset.metadata.orientation
        elif not to_pitch_dimensions:
            to_pitch_dimensions = dataset.metadata.pitch_dimensions

        if to_orientation == Orientation.BALL_OWNING_TEAM:
            if not dataset.metadata.flags & DatasetFlag.BALL_OWNING_TEAM:
                raise ValueError(
                    "Cannot transform to BALL_OWNING_TEAM orientation when dataset doesn't contain "
                    "ball owning team data"
                )

        transformer = cls(
            from_pitch_dimensions=dataset.metadata.pitch_dimensions,
            from_orientation=dataset.metadata.orientation,
            to_pitch_dimensions=to_pitch_dimensions,
            to_orientation=to_orientation,
        )
        metadata = replace(
            dataset.metadata,
            pitch_dimensions=to_pitch_dimensions,
            orientation=to_orientation,
        )
        if isinstance(dataset, TrackingDataset):
            frames = [
                transformer.transform_frame(record)
                for record in dataset.records
            ]

            return TrackingDataset(
                metadata=metadata,
                records=frames,
            )
        elif isinstance(dataset, EventDataset):
            events = list(map(transformer.transform_event, dataset.records))

            return EventDataset(
                metadata=metadata,
                records=events,
            )
        else:
            raise Exception("Unknown Dataset type")



from dataclasses import dataclass
from itertools import product
from types import MappingProxyType
from typing import (
    Dict,
    Generic,
    Iterator,
    List,
    Mapping,
    Sequence,
    Text,
    Tuple,
)

import networkx as nx

# noinspection PyProtectedMember
from .ast import (
    Alternation,
    AnyNumber,
    Capture,
    Concatenation,
    Final,
    Maybe,
    Node,
    _Initial,
    _Terminal,
)
from .matchers import Out, Tok, _TrailItem


def ast_to_graph(root: Node) -> nx.DiGraph:
    """
    You will create your regular expression with a specific syntax which is
    transformed into an AST, however the regular expression engine expects
    to navigate in a graph. As it is too complicated to navigate inside the
    AST directly, this function transforms the AST into an actual graph.

    Notes
    -----
    Since each node, except the :py:class:`nsre.ast.Final` ones, have children
    nodes, the idea is to insert nodes into the graph one at a time and then to
    work on those new nodes to transform it into its content.

    Also, there is implicitly a :code:`_Initial` and a :code:`_Terminal` node.
    The graph exploration will start from the initial node and the regular
    expression will be considered to be a match if when the input sequence is
    entirely consumed you can transition to the terminal node.

    By example, you got a node A which is a concatenation of B and C. Suppose
    that the code looks like this:

    >>> from nsre import *
    >>> c = Final(Eq('c'))
    >>> b = Final(Eq('b'))
    >>> a = c + b
    >>> g = ast_to_graph(a)

    Then the first graph you're going to get is

        :code:`_Initial` -> :code:`A` -> :code:`_Terminal`

    But then the algorithm is going to transform A into its content and you'll
    end up with the new graph

        :code:`_Initial` -> :code:`B` -> :code:`C` -> :code:`_Terminal`

    And so on if B and C have content of their own (they don't in the current
    example).

    The way to transform a node into its content depends on the node type, of
    course. That's why you'll find in this file a bunch of :code:`_explore_*`
    methods which are actually the ways to transform a specific node into a
    graph.

    The overall algorithm here is to have a to-do list (the "explore" variable)
    which contains the set of currently unexplored nodes. When a node is
    transformed into its content, the newly inserted nodes are also added to
    the to-do list and will be explored at the next iteration. This goes on
    and on until the whole AST has been transformed into a graph.

    Another detail is that capture groups are indicated with a start and a
    stop marker on the edges. Each edge can potentially contain in its data
    a "start_captures" or a "stop_captures" list. They contain the name, in
    order, of capture groups to start or stop. The capture should start right
    after the start and before the stop marker.

    See Also
    --------
    _explore_concatenation, _explore_alternation, _explore_maybe,
    _explore_any_number, _explore_capture
    """

    g = nx.DiGraph()
    initial = _Initial()
    terminal = _Terminal()

    g.add_nodes_from([initial, root, terminal])
    g.add_edge(initial, root)
    g.add_edge(root, terminal)

    explore = {root}

    while explore:
        for node in [*explore]:
            explore.remove(node)

            if isinstance(node, Final):
                pass
            elif isinstance(node, Concatenation):
                _explore_concatenation(explore, g, node)
            elif isinstance(node, Alternation):
                _explore_alternation(explore, g, node)
            elif isinstance(node, Maybe):
                _explore_maybe(explore, g, node)
            elif isinstance(node, AnyNumber):
                _explore_any_number(explore, g, node)
            elif isinstance(node, Capture):
                _explore_capture(explore, g, node)

    return g


def _explore_capture(explore, g, node):
    """
    Adds the capture flags on the edges around the node. Of another capture
    is already flagged, append to the list.
    """

    explore.add(node.statement)
    g.add_node(node.statement)

    for p in g.predecessors(node):
        data = g.get_edge_data(p, node, default={"start_captures": []})
        data["start_captures"] = [*data.get("start_captures", []), node]
        g.add_edge(p, node.statement, **data)

    for s in g.successors(node):
        data = g.get_edge_data(node, s, default={"stop_captures": []})
        data["stop_captures"] = [node, *data.get("stop_captures", [])]
        g.add_edge(node.statement, s, **data)

    g.remove_node(node)


# noinspection DuplicatedCode
def _explore_any_number(explore, g, node):
    """
    Allows to see any number of the child node. This means:

    - Either 0 match, meaning that all incoming edges are connected directly
      to all outgoing edges
    - Either 1 or more matches, granting a loop of the node on itself
    - And of course an exit edge which goes into the next node(s)
    """

    explore.add(node.statement)
    g.add_node(node.statement)

    g.add_edge(node.statement, node.statement)

    _cross_connect(g, node)

    for p in g.predecessors(node):
        data = g.get_edge_data(p, node, default={})
        g.add_edge(p, node.statement, **data)

    for s in g.successors(node):
        data = g.get_edge_data(node, s, default={})
        g.add_edge(node.statement, s, **data)

    g.remove_node(node)


# noinspection DuplicatedCode
def _explore_maybe(explore, g, node):
    """
    Like `_explore_any_number` but without the self-loop to allow at most one
    match.
    """

    explore.add(node.statement)
    g.add_node(node.statement)

    _cross_connect(g, node)

    for p in g.predecessors(node):
        data = g.get_edge_data(p, node, default={})
        g.add_edge(p, node.statement, **data)

    for s in g.successors(node):
        data = g.get_edge_data(node, s, default={})
        g.add_edge(node.statement, s, **data)

    g.remove_node(node)


def _cross_connect(g, node):
    """
    Used by `_explore_any_number` and `_explore_maybe` which both need to
    connect all incoming edges to all outgoing edges.

    Notes
    -----
    There is a complicated logic with the capture flags in order to make sure
    that capture groups that are opened and closed at the same time don't
    actually stay (because they confuse the matching algorithm and they are
    useless).
    """

    for p, s in product(g.predecessors(node), g.successors(node)):
        data1 = g.get_edge_data(p, node, default={})
        data2 = g.get_edge_data(node, s, default={})
        merged = dict(**data1)
        merged.update(**data2)

        if "start_captures" in data1 and "start_captures" in data2:
            merged["start_captures"] = (
                data1["start_captures"] + data2["start_captures"]
            )

        if "stop_captures" in data1 and "stop_captures" in data2:
            merged["stop_captures"] = (
                data1["stop_captures"] + data2["stop_captures"]
            )

        cancel = 0
        start_captures = merged.get("start_captures", [])
        stop_captures = merged.get("stop_captures", [])

        for start, stop in zip(start_captures, stop_captures):
            if start == stop:
                cancel += 1
            else:
                break

        if cancel:
            merged["start_captures"] = start_captures[cancel:]
            merged["stop_captures"] = stop_captures[:-cancel]

        g.add_edge(p, s, **merged)


def _explore_alternation(explore, g, node):
    """
    This node accepts either the left or the right child node. Meaning that
    all edges connected to this node are connected to both the left and the
    right node.
    """

    explore.add(node.left)
    explore.add(node.right)

    g.add_node(node.left)
    g.add_node(node.right)

    for p in g.predecessors(node):
        data = g.get_edge_data(p, node, default={})
        g.add_edge(p, node.left, **data)
        g.add_edge(p, node.right, **data)

    for s in g.successors(node):
        data = g.get_edge_data(node, s, default={})
        g.add_edge(node.left, s, **data)
        g.add_edge(node.right, s, **data)

    g.remove_node(node)


# noinspection DuplicatedCode
def _explore_concatenation(explore, g, node):
    """
    Concatenation is just taking all the incoming edges and plugging them into
    the left node, making a connection between the left and the right node and
    then taking all the outgoing edges and plugging them into the right node.
    """

    explore.add(node.left)
    explore.add(node.right)

    g.add_node(node.left)
    g.add_node(node.right)

    for p in g.predecessors(node):
        data = g.get_edge_data(p, node, default={})
        g.add_edge(p, node.left, **data)

    g.add_edge(node.left, node.right)

    for s in g.successors(node):
        data = g.get_edge_data(node, s, default={})
        g.add_edge(node.right, s, **data)

    g.remove_node(node)


@dataclass(frozen=True)
class Explorer(Generic[Tok, Out]):
    """
    An explorer is a pointer to a specific position in the graph, with a past
    trail of previously visited nodes.
    """

    re: "RegExp[Tok, Out]"
    node: Node[Tok, Out]
    trail: Tuple[_TrailItem[Out], ...]

    @property
    def signature(self) -> Tuple[Node, Tuple[_TrailItem, ...]]:
        """
        Sortable signature for this explorer, used for de-duplication
        """

        return self.node, self.trail

    def advance(self, token: Tok) -> Iterator["Explorer[Tok, Out]"]:
        """
        Given the provided token, emits all the explorers that managed to
        advance to another node.

        Parameters
        ----------
        token
            Consumed token
        """

        for s in self.re.graph.successors(self.node):
            if not isinstance(s, Final):
                continue

            data = self.re.graph.get_edge_data(self.node, s, default={})
            possible_trail = self.trail + (_TrailItem(item=None, data=data),)

            for m in s.statement.match(token, trail=possible_trail):
                yield Explorer(
                    re=self.re,
                    node=s,
                    trail=self.trail + (_TrailItem(item=m, data=data),),
                )

    def can_terminate(self) -> bool:
        """
        Indicates if this explorer is connected to a _Terminal node, meaning
        that if you were to stop the matching here it would mean that the
        expression matched.
        """
        return self.re.graph.has_successor(self.node, _Terminal())


class _Match(Generic[Out]):
    """
    Internal sub-type that helps building the Match and MatchList objects
    expected outside of `RegExp#match()`.
    """

    def __init__(self, start_pos: int):
        self.start_pos = start_pos
        self.children: Dict[Text, List[_Match]] = {}
        self.trail: List[Out] = []
        self._stack: List[Capture] = []

    def _deep_get(self, keys: Sequence[Capture]) -> "_Match":
        """
        Gets a child following a path of keys. At each level, gets the last
        child in the list as it is the one that we're going to want to deal
        with.

        Parameters
        ----------
        keys
            Successive name of keys to follow
        """

        ptr = self

        for key in keys:
            ptr = ptr.children[key.name][-1]

        return ptr

    def start(self, capture: Capture, pos: int) -> None:
        """
        Call this to indicate that you're starting a capture group

        Notes
        -----
        This will append the name of the capture group to the capture stack
        and will also make sure that there is a corresponding child Match for
        this sequence of keys.

        Parameters
        ----------
        capture
            Name of the capturing group
        pos
            Starting index in the input sequence
        """

        stick = self._deep_get(self._stack)

        if capture not in stick.children:
            stick.children[capture.name] = [_Match(pos)]
        else:
            stick.children[capture.name].append(_Match(pos))

        self._stack.append(capture)

    def stop(self, capture: Capture) -> None:
        """
        Indicates that matching should stop (as opposed to start).

        Parameters
        ----------
        capture
            Name of the matching group to stop
        """

        if capture != self._stack[-1]:
            raise ValueError("Trying to stop a match which is not started")

        self._stack = self._stack[0:-1]

    def append(self, item: Out) -> None:
        """
        Appends the item at all level of trails for the current stack.

        Parameters
        ----------
        item
            Matched item to be added to trails
        """

        self.trail.append(item)
        ptr = self

        for key in self._stack:
            ptr = ptr.children[key.name][-1]
            ptr.trail.append(item)

    def as_match(self, join_trails: bool = False) -> "Match":
        """
        Converts this into a real read-only match object.

        Parameters
        ----------
        join_trails
            If set to true then the items matched in the trail are expected to
            be individual characters and they will be joined in a string
            instead of being returned in an array
        """

        return Match(
            start_pos=self.start_pos,
            trail="".join(self.trail) if join_trails else tuple(self.trail),
            children=MappingProxyType(
                {
                    k: MatchList(i.as_match(join_trails) for i in v)
                    for k, v in self.children.items()
                }
            ),
        )


@dataclass(frozen=True)
class Match(Generic[Out]):
    """
    Represents a match for a capture group in the regular expression.
    """

    # Index of the first matching token
    start_pos: int

    # Sub-groups that matched. As several groups could be matching, a list
    # is returned so you can access each one of them.
    children: Mapping[Text, List["Match"]]

    # Trail of matched items. It's the output of the matcher, not the input
    # tokens.
    trail: Sequence[Out]

    def __getitem__(self, item):
        return self.children[item][0]


class MatchList(tuple, Generic[Out]):
    """
    List of matches. It's just a convenience around a tuple in order to
    facilitate getting a specific group in the first item of the list.
    """

    def __getitem__(self, item) -> Match[Out]:
        if isinstance(item, str):
            return self[0][item]
        else:
            return super().__getitem__(item)


def _make_match(trail: Tuple[_TrailItem, ...]) -> _Match[Out]:
    """
    Transforms an explorer into a Match object using its trail

    Parameters
    ----------
    explorer
        Explorer that you want to transform
    """

    match = _Match(0)

    for i, token in enumerate(trail):
        for stop in token.data.get("stop_captures", []):
            match.stop(stop)

        for start in token.data.get("start_captures", []):
            match.start(start, i)

        match.append(token.item)

    return match


class RegExp(Generic[Tok, Out]):
    """
    Core of the RegExp system. Don't instantiate this directly. There is so
    far only one way to create an instance easily but in the future there will
    be more ways, like a parser for good old-school regular expressions or a
    new specific DSL.

    Here's an usage example.

    >>> from nsre import *
    >>> re = RegExp.from_ast(anything()['user'] + seq("@") + anything()['domain'])
    >>> m = re.match('remy.sanchez@with-madrid.com', join_trails=True)
    >>> assert m['user'].trail == 'remy.sanchez'
    >>> assert m['domain'].trail == 'with-madrid.com'
    """

    def __init__(self, graph: nx.DiGraph):
        """
        Don't call me directly.

        See Also
        -------
        RegExp#from_ast() : compiles an AST into a regular expression

        Parameters
        ----------
        graph
            The regular expression's graph
        """

        self.graph = graph

    @classmethod
    def from_ast(cls, root: Node[Tok, Out]) -> "RegExp[Tok, Out]":
        """
        Use this to generate your regular expression. To generate the AST,
        have a look at :py:mod:`nsre.ast` and :py:mod:`nsre.shortcuts` modules.

        Parameters
        ----------
        root
            Root node of your expression.
        """

        return cls(graph=ast_to_graph(root.copy()))

    def match(
        self,
        seq: Sequence[Tok],
        join_trails: bool = False,
        consume_all: bool = True,
    ) -> MatchList[Match[Out]]:
        """
        For a given sequence of tokens, generates all the matches that were
        detected.

        Notes
        -----
        If you think your regular expression like you would think of using the
        `re` module then you're going to have 0 or 1 match. However, if your
        Matchers can match several options then you might end up with two
        or more matches at the same time.

        For this reason, the MatchList object provides shortcuts so you don't
        have to skim through the list of matches if you don't want to.

        Please note that inside a match, the capture groups do not match in
        "parallel", only root MatchList provides several matching options. The
        inside of them is just their content.

        Parameters
        ----------
        seq
            Sequence that you would like to test
        join_trails
            If all your output items are going to be characters, you can set
            this to true in order to receive trails that are strings instead of
            them being character lists.
        """

        stack: List[Explorer[Tok, Out]] = [Explorer(self, _Initial(), tuple())]

        stacks = []
        for token in seq:
            stack = list(
                self._de_duplicate(
                    ne for oe in stack for ne in oe.advance(token)
                )
            )

            # if not consume_all:
            #     if any(s.can_terminate() for s in stack):
            #         break

            if not stack:
                break

            stacks.append(stack)

        if not consume_all and stacks:
            stacks.reverse()
            for stack in stacks:
                stack = [s for s in stack if s.can_terminate()]
                if stack:
                    break

        terminal = list(
            self._de_duplicate(
                (s for s in stack if s.can_terminate()), key="trail"
            )
        )

        return MatchList(
            _make_match(s.trail).as_match(join_trails=join_trails)
            for s in terminal
        )

    def _de_duplicate(
        self, stack: Iterator[Explorer[Tok, Out]], key: Text = "signature"
    ) -> Iterator[Explorer[Tok, Out]]:
        """
        As there is potentially several paths that lead to the same result, we
        merge for each node all identical trails. Without this the number of
        results becomes completely crazy (on top of being useless and
        confusing)
        """

        stack = list(sorted(stack, key=lambda e: getattr(e, key)))

        if not stack:
            return

        yield stack[0]

        for i in range(1, len(stack)):
            if getattr(stack[i], key) != getattr(stack[i - 1], key):
                yield stack[i]


__all__ = ["RegExp", "Match", "MatchList", "ast_to_graph", "_make_match"]



import os, logging

import requests

from typing import Dict, Union

from kloppy.domain import TrackingDataset, EventDataset

from .registered import _DATASET_REGISTRY


logger = logging.getLogger(__name__)


def download_file(url, local_filename):
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)


def get_local_files(
    dataset_name: str, files: Dict[str, str]
) -> Dict[str, str]:
    datasets_base_dir = os.environ.get("KLOPPY_BASE_DIR", None)
    if not datasets_base_dir:
        datasets_base_dir = os.path.expanduser("~/kloppy_datasets")

    dataset_base_dir = f"{datasets_base_dir}/{dataset_name}"
    if not os.path.exists(dataset_base_dir):
        os.makedirs(dataset_base_dir)

    local_files = {}
    for file_key, file_url in files.items():
        filename = f"{file_key}={file_url.split('/')[-1]}"
        local_filename = f"{dataset_base_dir}/{filename}"
        if not os.path.exists(local_filename):
            logger.info(f"Downloading {filename}")
            download_file(file_url, local_filename)
            logger.info("Download complete")
        else:
            logger.info(f"Using local cached file {local_filename}")
        local_files[file_key] = local_filename
    return local_files


def load(
    dataset_name: str, options=None, **dataset_kwargs
) -> Union[TrackingDataset, EventDataset]:
    if dataset_name not in _DATASET_REGISTRY:
        raise ValueError(
            f"Dataset {dataset_name} not found. Known datasets: {', '.join(_DATASET_REGISTRY.keys())}"
        )

    builder_cls = _DATASET_REGISTRY[dataset_name]
    builder = builder_cls()

    dataset_urls = builder.get_dataset_urls(**dataset_kwargs)
    dataset_local_files = get_local_files(dataset_name, dataset_urls)

    file_handlers = {
        local_file_key: open(local_file_name, "rb")
        for local_file_key, local_file_name in dataset_local_files.items()
    }

    try:
        serializer_cls = builder.get_serializer_cls()
        serializer = serializer_cls()
        dataset = serializer.deserialize(inputs=file_handlers, options=options)
    finally:
        for fp in file_handlers.values():
            fp.close()
    return dataset



from dataclasses import replace

from kloppy.domain import List, EventDataset

# register all of them
from . import builders

from .registered import create_state_builder


def add_state(dataset: EventDataset, *builder_keys: List[str]) -> EventDataset:
    """
    Add state

    Arguments:
        - builder_keys: `lineup` `score` `sequence`

    Examples:
        >>> dataset = dataset.add_state('lineup', 'score')

    Returns:
        [`EventDataset`][kloppy.domain.models.event.EventDataset]
    """
    if len(builder_keys) == 1 and isinstance(builder_keys[0], list):
        builder_keys = builder_keys[0]

    builders = {
        builder_key: create_state_builder(builder_key)
        for builder_key in builder_keys
    }

    state = {
        builder_key: builder.initial_state(dataset)
        for builder_key, builder in builders.items()
    }

    events = []
    for event in dataset.events:

        state = {
            builder_key: builder.reduce_before(state[builder_key], event)
            for builder_key, builder in builders.items()
        }

        events.append(replace(event, state=state))

        state = {
            builder_key: builder.reduce_after(state[builder_key], event)
            for builder_key, builder in builders.items()
        }

    return replace(dataset, records=events)



from abc import ABC, abstractmethod
from typing import Tuple, Dict

from kloppy.utils import Readable
from kloppy.domain import CodeDataset


class CodeDataSerializer(ABC):
    @abstractmethod
    def deserialize(
        self, inputs: Dict[str, Readable], options: Dict = None
    ) -> CodeDataset:
        raise NotImplementedError

    @abstractmethod
    def serialize(self, dataset: CodeDataset) -> bytes:
        raise NotImplementedError



from abc import abstractmethod
from typing import TypeVar

from kloppy.domain import EventDataset, Event
from .registered import RegisteredStateBuilder

T = TypeVar("T")


class StateBuilder(metaclass=RegisteredStateBuilder):
    @abstractmethod
    def initial_state(self, dataset: EventDataset) -> T:
        pass

    @abstractmethod
    def reduce_before(self, state: T, event: Event) -> T:
        pass

    @abstractmethod
    def reduce_after(self, state: T, event: Event) -> T:
        pass
